// Copyright (C) 2017-2020  Nexedi SA and Contributors.
//                          Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

package neo
// infrastructure for creating NEO test clusters.

import (
	"context"
	"fmt"
	"sync"

	"lab.nexedi.com/kirr/go123/xnet"
	"lab.nexedi.com/kirr/go123/xnet/pipenet"
	"lab.nexedi.com/kirr/go123/xsync"

	"lab.nexedi.com/kirr/neo/go/internal/xcontext"
	"lab.nexedi.com/kirr/neo/go/internal/xtracing/tracetest"
	"lab.nexedi.com/kirr/neo/go/neo/proto"

	"lab.nexedi.com/kirr/neo/go/neo/storage"
	"lab.nexedi.com/kirr/neo/go/zodb"
)


// tCluster is a test NEO cluster.
//
// Create it with tNewCluster.
// Create nodes with .NewMaster, .NewStorage and .NewClient.
//
// XXX text about events tracing
type tCluster struct {
	*tracetest.T // original testing env this cluster was created at

	name	string           // name of the cluster
	network	*pipenet.Network // nodes interoperate via netowrk XXX -> lo

	gotracer *TraceCollector // for tracing go nodes XXX -> GoTracer
	//tpy	 *PyTracer       // for tracing py nodes
	erouter  *EventRouter    // to which stream an event should go

	tabMu     sync.Mutex
	nodeTab	  map[string/*node*/]*tNode

	runWG     *xsync.WorkGroup // started nodes are .Run under runWG
	runCancel func()
}

// tNode represents information about a test node ... XXX
type tNode struct {
	net xnet.Networker
}

// XXX stub
type ITestMaster interface {
//	Run(ctx context.Context) error
	Start() error
}

type ITestStorage interface {
//	Run(ctx context.Context) error
}

type ITestClient interface {
//	Run(ctx context.Context) error

	zodb.IStorageDriver
}

// tNewCluster creates new NEO test cluster.
//
// At the end of the test the cluster has to be stopped via t.Stop().
//
// TODO add options describing whether node type X should be created via Go or via Py.
func tNewCluster(ttest *tracetest.T, name string) *tCluster {
	t := &tCluster{
		T:    ttest,
		name: name,

		network: pipenet.New("testnet"),	// test network

		nodeTab: make(map[string]*tNode),
	}

	t.erouter  = NewEventRouter()
	t.gotracer = NewTraceCollector(ttest)
	ttest.SetEventRouter(t.erouter.RouteEvent)

	t.gotracer.Attach()

	runCtx, runCancel := context.WithCancel(context.Background())
	t.runWG = xsync.NewWorkGroup(runCtx)
	t.runCancel = runCancel

	return t
}

// Stop stops the cluster.
//
// After Stop completes all processes of the cluster are stopped.
func (t *tCluster) Stop() {
	t.Helper()
	t.TraceOff()

	t.runCancel()
	err := t.runWG.Wait()
	if xcontext.Canceled(err) { // cancel is expected
		err = nil
	}
	// XXX t.network.Close()
	if err != nil {
		t.Fatalf("run shutdown: %s", err)
	}
}

// TraceOff disables tracing.
// Trace events will stop to be created and delivered.
func (t *tCluster) TraceOff() {
	t.gotracer.Detach()
	//XXX t.pytracer.Detach()
}

// Checker returns tracechecker corresponding to name.
//
// name might be of "node" or "node1-node2" kind.	XXX more text
// node or node1/node2 must be already registered.
//
// XXX is this a good idea?
// XXX move into tracetest.T?  (e.g. T.Stream())
func (t *tCluster) Checker(name string) *tStreamChecker {
	return &tStreamChecker{t: t, stream: name}
}
type tStreamChecker struct {
	t      *tCluster
	stream string
}
func (c *tStreamChecker) Expect(event interface{}) {
	c.t.Helper()
	c.t.Expect(c.stream, event)
}
/*
func (t *tCluster) Checker(name string) *tracetest.EventChecker {
	t.tabMu.Lock()
	defer t.tabMu.Unlock()

	c, ok := t.checkTab[name]
	if !ok {
		panic(fmt.Sprintf("test cluster: no checker for %q", name))
	}

	return c
}
*/

// registerNewNode registers new node with given name.
//
// the node is registered in .nodeTab and .checkTab is populated ... XXX
//
// the node must not be registered before.
func (t *tCluster) registerNewNode(name string) *tNode {
	t.tabMu.Lock()
	defer t.tabMu.Unlock()

	// check node not yet registered
	if _, already := t.nodeTab[name]; already {
		panic(fmt.Sprintf("test cluster: node %q registered twice", name))
	}

	// trace of events local to node
	t.erouter.BranchNode(name, name)

	// trace for events on links of all node1-node2 pairs
	for name2 := range t.nodeTab {
		// traces for events of 2 streams:
		// 1) with cause root being node1 -> node2 send, and
		// 2) ----//----            node1 <- node2 send.
		t.erouter.BranchLink(name + "-" + name2, name+"-"+name2, name2+"-"+name)
	}

	node := &tNode{}
	node.net  = xnet.NetTrace(t.network.Host(name), t.gotracer)

	t.nodeTab[name] = node
	return node
}


// NewMaster creates new master on node name.
//
// The master will be accepting incoming connections at node:1.
// The node must be not yet existing and will be dedicated to the created master fully.	XXX
//
// XXX error of creating py process? -> t.Fatal
func (t *tCluster) NewMaster(name string) ITestMaster {
	node := t.registerNewNode(name)
	m := tNewMaster(t.name, ":1", node.net)

	// let tracer know how to map state addresses to node names
	t.gotracer.RegisterNode(m.node, name)

	// start the node
	t.runWG.Go(func(ctx context.Context) error {
		return m.Run(ctx)
	})

	return m
}

func (t *tCluster) NewStorage(name, masterAddr string, back storage.Backend) ITestStorage {
	node := t.registerNewNode(name)
	s := tNewStorage(t.name, masterAddr, ":1", node.net, back)
	t.gotracer.RegisterNode(s.node, name)
	t.runWG.Go(func(ctx context.Context) error {
		return s.Run(ctx)
	})
	return s
}

func (t *tCluster) NewClient(name, masterAddr string) ITestClient {
	node := t.registerNewNode(name)
	c := NewClient(t.name, masterAddr, node.net)
	t.gotracer.RegisterNode(c.node, name)
	t.runWG.Go(func(ctx context.Context) error {
		return c.Run(ctx)
	})
	return c
}



// tStorage is test-wrapper around Storage.
//
// - to automatically listen by address, not provided listener.
type tStorage struct {
	*Storage
	serveAddr string
}

func tNewStorage(clusterName, masterAddr, serveAddr string, net xnet.Networker, back storage.Backend) *tStorage {
	return &tStorage{
		Storage:   NewStorage(clusterName, masterAddr, net, back),
		serveAddr: serveAddr,
	}
}

func (s *tStorage) Run(ctx context.Context) error {
	l, err := s.node.Net.Listen(ctx, s.serveAddr)
	if err != nil {
		return err
	}

	return s.Storage.Run(ctx, l)
}

// tMaster is test-wrapper around Master.
//
// - automatically listens by address, not provided listener.
// - uses virtual clock.
type tMaster struct {
	*Master
	serveAddr string
	vclock    vclock
}

func tNewMaster(clusterName, serveAddr string, net xnet.Networker) *tMaster {
	m := &tMaster{
		Master:   NewMaster(clusterName, net),
		serveAddr: serveAddr,
	}
	m.Master.monotime = m.vclock.monotime
	return m
}

func (m *tMaster) Run(ctx context.Context) error {
	l, err := m.node.Net.Listen(ctx, m.serveAddr)
	if err != nil {
		return err
	}

	return m.Master.Run(ctx, l)
}


// --------

// tStartSimpleNEOGoSrv starts simple NEO/go server with 1 master and 1 storage.
// The cluster is returned in ready-to-start.
func tStartSimpleNEOGoSrv(name string, Sback storage.Backend) *tCluster {
	t := tNewCluster(name)
	M := t.NewMaster("m")

	// XXX if we would always use lonet or pipenet - all network addresses would be predictable

	// M starts listening - determine its listen address
	t.Expect("m", netlisten("xxx")) // XXX xxx will be the address
	t.Expect("m", δnode("m", Maddr, proto.MASTER, 1, proto.RUNNING, proto.IdTimeNone))
	t.Expect("m", clusterState("m", proto.ClusterRecovering))

	S := t.NewStorage("s", Maddr, Sback)

	// S starts listening
	t.Expect("s", netlisten("yyy")) // XXX yyy will be the address

	// S connects to M
	t.Expect("s-m", netdial("s", Maddr))
	t.Expect("s-m", netconnect("sXXX", "mYYY",  Maddr))
	t.Expect("s-m", conntx("s:2", "m:2", 1, &proto.RequestIdentification{
		NodeType:	proto.STORAGE,
		UUID:		0,
		Address:	xnaddr("s:1"),
		ClusterName:	"abc1",
		IdTime:		proto.IdTimeNone,
		DevPath:        nil,
		NewNID:         nil,
	}))

	t.Expect("m", δnode("m", "s:1", proto.STORAGE, 1, proto.PENDING, 0.01))

	t.Expect("s-m", conntx("m:2", "s:2", 1, &proto.AcceptIdentification{
		NodeType:	proto.MASTER,
		MyUUID:		proto.UUID(proto.MASTER, 1),
		YourUUID:	proto.UUID(proto.STORAGE, 1),
	}))

	// M starts recovery on S
	t.Expect("m-s", conntx("m:2", "s:2", 0, &proto.Recovery{}))
	t.Expect("m-s", conntx("s:2", "m:2", 0, &proto.AnswerRecovery{
		// empty new node
		PTid:		0,
		BackupTid:	proto.INVALID_TID,
		TruncateTid:	proto.INVALID_TID,
	}))

	t.Expect("m-s", conntx("m:2", "s:2", 2, &proto.AskPartitionTable{}))
	t.Expect("m-s", conntx("s:2", "m:2", 2, &proto.AnswerPartitionTable{
		PTid:		0,
		NumReplicas:	0,
		RowList:	[]proto.RowInfo{},
	}))

	// M ready to start: new cluster, no in-progress S recovery
	t.Expect("m", masterStartReady("m", true))

	err := M.Start()
	if err != nil {
		t.Fatal(err)
	}

	return t
}
